import abc


class LoggingInterface:
    """Contract for logging.

    Attributes:
        debug_on (bool): Debug status, default `False`.
    """
    debug_on = False

    @abc.abstractmethod
    def debug(self, msg: str):
        pass

    @abc.abstractmethod
    def info(self, msg: str):
        pass

    @abc.abstractmethod
    def error(self, msg: str):
        pass



from unittest import TestCase
from constants_and_utils.constants.enums import (
    HttpCode,
    Gender
    )


class TestEnums(TestCase):

    def test_http_code(self):
        expected = [
            200, 201, 202, 204, 400,
            401, 402, 403, 404, 405,
            406, 500, 502, 504]

        result = HttpCode.values()

        self.assertEqual(result, expected)

    def test_gender(self):
        expected = ['F', 'M']

        result = Gender.values()

        self.assertEqual(result, expected)



from unittest import TestCase
from constants_and_utils.utils.use_cases import (
    Response,
    ExportResponse
)


class TestResponse(TestCase):

    def test_ok(self):
        expected = {'errors': None, 'http_code': 200, 'message': 'OK'}
        result = Response(
            http_code=200,
            message='OK'
            )

        self.assertEqual(result.__json__(), expected)


class TestExportResponse(TestCase):

    def test_ok(self):
        expected = {
            'errors': None, 'http_code': 200, 'message': 'OK',
            'file_name': 'file.csv'
            }

        result = ExportResponse(
            http_code=200,
            message='OK',
            file_name='file.csv'
            )

        self.assertEqual(result.__json__(), expected)



from unittest import TestCase
from constants_and_utils.utils.decorators import (
    result_to_json
    )


class TestDecorators(TestCase):

    def test_result_to_json_dict(self):
        expected = '{"message": "hi, people"}'

        @result_to_json
        def hello(name):
            return {'message': f'hi, {name}'}

        result = hello('people')

        self.assertEqual(result, expected)



from enum import Enum, unique


@unique
class HttpCode(Enum):
    """Http code constants.

    Examples:
        Use case.

        >>> HttpCode.OK.value
        200
    """
    # 2xx Success
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    NO_CONTENT = 204

    # 4xx Client Error
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    PAYMENT_REQUIRED = 402
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406

    # 5xx Server Error
    INTERNAL_SERVER_ERROR = 500
    BAD_GATEWAY = 502
    GATEWAY_TIMEOUT = 504

    @staticmethod
    def values() -> list:
        """returns list of values"""
        return list(map(lambda e: e.value, HttpCode))


@unique
class Gender(Enum):
    """Gender constants.

    Examples:
        Use case.

        >>> Gender.FEMALE.value
        'F'
    """
    FEMALE = 'F'
    MALE = 'M'

    @staticmethod
    def values() -> list:
        """returns list of values"""
        return list(map(lambda e: e.value, Gender))



import abc
from ..constants.enums import HttpCode
from .contract import LoggingInterface


class Response:
    """Response object for use cases.

    Attributes:
        _http_code (constants.enums.HttpCode):
            Guarantee valid code. Default HttpCode.OK.
        message (str): A message.
        errors (list): List of errors.

    Examples:
        Instantiate class.

        >>> r = Response(http_code=200, message='OK')
        >>> r.message
        'OK'
    """

    def __init__(
        self,
        http_code: int = None,
        message: str = None,
        errors: list = None
    ):
        # Default HttpCode
        self._http_code: HttpCode = HttpCode.OK

        self.http_code = http_code
        self.message = message
        self.errors = errors

    @property
    def http_code(self) -> int:
        """int: code http"""
        if self._http_code:
            return self._http_code.value
        return None

    @http_code.setter
    def http_code(self, http_code) -> None:
        """validate http code"""
        if isinstance(http_code, HttpCode):
            self._http_code = http_code
        elif http_code:
            self._http_code = HttpCode(http_code)

    def __json__(self):
        return {
            'http_code': self.http_code,
            'message': self.message,
            'errors': self.errors,
        }


class UseCaseInterface(metaclass=abc.ABCMeta):
    """Contract for use cases.

    Attributes:
        _path (str): A path.
        _logger (utils.contract.LoggingInterface): Logger class.
    """
    _path = None
    _logger: LoggingInterface = None

    def _debugger(self, message):
        """Run if the debug is actived.

        Args:
            message (str): A message.
        """
        self._logger.debug(
            '{}: {}'.format(self._path, message)
            )

    def _error(self, message):
        """Error log.

        Args:
            message (str): A message.
        """
        self._logger.error(
            '{}: {}'.format(self._path, message)
            )

    @abc.abstractmethod
    def handle(self, *args, **kwargs) -> Response:
        """Unique public methed, return use case result.

        Args:
            request (dict): (Optional) Recommended.
            args: Other positional arguments.
            kwargs: Other keyword arguments.

        Returns:
            Response: Response Instance or
                daughter class (recommended if you want to
                return data other than attributes of the reponse).
        """
        pass


class ExportResponse(Response):
    """Response for use cases export file.

    Attributes:
        _http_code (constants.enums.HttpCode):
            Guarantee valid code. Default HttpCode.OK.
        message (str): A message.
        errors (list): List of errors.
        file_name (str): A filename or path.

    Examples:
        Instantiate class.

        >>> r = Response(http_code=200, message='OK', file_name='file.csv')
        >>> r.file_name
        'file.csv'
    """
    def __init__(self, file_name: str = None, *args, **kwargs):
        self.file_name = file_name

        super().__init__(*args, **kwargs)

    def __json__(self):
        result = super().__json__()
        result['file_name'] = self.file_name
        return result



import json


def result_to_json(func):
    """Convert result into json from a function.

    Args:
        func: Function to execute.

    Returns:
        Result json dumps.

    Examples:
        Decorate function.

        >>> @result_to_json
        ... def hello():
        ...     return {'message': 'hi'}
    """
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, tuple) and isinstance(result[0], dict):
            result = list(result)
            result[0] = json.dumps(result[0])
            result = tuple(result)
        elif isinstance(result, dict):
            result = json.dumps(result)
        return result
    return wrapper


def debugger_handle(func):
    """Use debugger_handle decorator to show received data and response data.

    Args:
        func: Function to execute.

    Returns:
        Result handle.

    Examples:
        Decorate function.

        >>> @debugger_handle
        ... def handle():
        ...     return {'message': 'hi'}
    """

    def wrapper(self, *args, **kwargs):
        self._debugger('kwargs: {}'.format(kwargs))
        self._debugger('args: {}'.format(args))
        result = func(self, *args, **kwargs)
        self._debugger(result.__json__())
        return result
    wrapper.__doc__ = func.__doc__
    return wrapper



from functools import reduce
import re


def recursive_get(d: dict, keys, default=None):
    """Apply a get recursive to the dictionary.

    Args:
        d (dict): Dictionary.
        keys (str|list|tuple): If is a str keyA.keyB => [keyA, keyB].
        default (mixed): Not required.

    Returns:
        The return value.

    Examples:
        Run function.

        >>> recursive_get({'NUMBER': {'ONE': 1 }}, 'NUMBER.ONE')
        1
    """
    if isinstance(keys, str):
        keys = keys.split('.')
    result = reduce(lambda c, k: c.get(k, {}), keys, d)
    if default is not None and result == {}:
        return default
    return result


def kebab_case(string: str) -> str:
    """Convert string into kebab case or slug.

    Args:
        string (str): String to convert.

    Returns:
        str: String en kebab-case.

    Examples:
        Run function.

        >>> kebab_case('Hello world')
        'hello-world'
    """
    components = string.split(' ')
    string = '-'.join([str(elem) for elem in components])
    return string.lower()


def camel_to_snake(string: str):
    """Convert a string from camel case to snake case.

    Args:
        string (str): String to convert.

    Returns:
        str: String value.

    Examples:
        Run function.

        >>> camel_to_snake('HelloWorld')
        'hello_world'
    """
    string = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', string).lower()


def dict_keys_to_snake(d: dict) -> dict:
    """Convert the keys from camel case dictionary to snake case.

    Args:
        d (dict): dictionay to convert.

    Returns:
        dict: dictionary converted.

    Examples:
        Run function.

        >>> d = {'HelloWorld': 'message'}
        >>> dict_keys_to_snake(d)
        {'hello_world': 'message'}
    """
    return dict(
        zip([camel_to_snake(k) for k in d.keys()], list(d.values()))
    )



from unittest import TestCase
from constants_and_utils.utils.helpers import (
    recursive_get,
    kebab_case,
    camel_to_snake,
    dict_keys_to_snake
)


class TestHelpers(TestCase):

    def test_recursive_get(self):
        param_d = {'NUMBER': {'ONE': 1}}
        param_keys = 'NUMBER.ONE'
        expected = 1

        result = recursive_get(
            d=param_d,
            keys=param_keys
        )

        self.assertEqual(result, expected)

    def test_kebab_case(self):
        param_str = 'Hello world'
        expected = 'hello-world'

        result = kebab_case(
            string=param_str
        )

        self.assertEqual(result, expected)

    def test_camel_to_snake(self):
        param_str = 'HelloWorld'
        expected = 'hello_world'

        result = camel_to_snake(
            string=param_str
        )

        self.assertEqual(result, expected)

    def test_dict_keys_to_snake(self):
        param_dict = {'HelloWorld': 'message'}
        expected = {'hello_world': 'message'}

        result = dict_keys_to_snake(
            d=param_dict
        )

        self.assertEqual(result, expected)
